#!/usr/bin/perl

#created by Matthew Hestand, Michiel van Galen, Michel Villerius

#Copyright 2010 Leiden University Medical Center http://www.lgtc.nl/GAPSS


#This file is part of GAPSS.
#
#    GAPSS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    GAPSS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with GAPSS.  If not, see <http://www.gnu.org/licenses/>.

#===============================================#
#	IntermediateGapssFormat2WIGGLE		#
#===============================================#
#       Created by 	m.van_galen@lumc	#
#       		August, 2009		#
#===============================================#

# Parse igf formatted file and create wiggle tracks accordingly
# Checks for existence of reserved filenames and check input formats

use strict;
use Cwd;

#======================
# Check input parameters
#======================
if (@ARGV < 4){
	if (($ARGV[0] eq "--help") or (!$ARGV[0])){
		print "\nUSAGE: $0 -i <STRING> -s <BOOL> -c <INT>
		
	Options:
	-i	REQ	IGF file 
	-s	OPT	Seperate strands [1=yes,0=no] DEFAULT=0
	-c	OPT	Cutoff for wig-file. Depths below this value are not printed. Zeros are never printed.	DEFAULT=1
\n";	
		exit(1);		
	}
	else{
		print "\nFor help on how to run this tool:\n\n$0 --help\n\n";
		exit(1);
	}
}
my %params;
my %seen_args;
my %used_args;
foreach my $i (0..$#ARGV){
	$seen_args{$ARGV[$i]}++;
	if($ARGV[$i] eq "-i"){
		if($ARGV[$i+1]){
			$params{"in_file"}=$ARGV[$i+1];
			$used_args{$ARGV[$i]}++;
			$used_args{$ARGV[$i+1]}++;			
		}
		else{
			print STDERR "Please enter valid input filename ...\n";
			exit(2);     
		}
	}
	elsif($ARGV[$i] eq "-s"){
		if($ARGV[$i+1]==0 || $ARGV[$i+1]==1 ){
			$params{"strand"}=$ARGV[$i+1];
			$used_args{$ARGV[$i]}++;
			$used_args{$ARGV[$i+1]}++;			
		}else{
			print STDERR "Please enter valid argument for separation ...\n";
			exit(2);     
		}
	}
	elsif($ARGV[$i] eq "-c"){
		if($ARGV[$i+1]){
			$params{"cutoff"}=$ARGV[$i+1];
			$used_args{$ARGV[$i]}++;
			$used_args{$ARGV[$i+1]}++;			
		}else{
			print STDERR "Please enter valid cutoff ...\n";
			exit(2);     
		}
	}	
}
foreach my $seen_this_arg(keys %seen_args){if(!exists $used_args{$seen_this_arg}){die("Error, encountered unused/invalid argument $seen_this_arg\n");}}
# Check if input file is supplied and existence
if(!exists($params{"in_file"})){die("No input file submitted ...\n");}
if(!-e $params{"in_file"}){die("Input file ".$params{"in_file"}." not found ...\n")}
# Check strand seperation arguments
# if(!exists($params{"strand"})){die("No strand separation choice submitted ...\n");}
if($params{"strand"}){if($params{"strand"} ne "0" && $params{"strand"} ne "1"){die ("Incorrect argument for -s please use [1 or 0] ...\n");}}
else{$params{"strand"}=0;}
# Check cutoff for wigtrack
if($params{"cutoff"}){if($params{"cutoff"} =~ m/\D/gi){die ("Please enter a numeric value for wig-file cutoff ...\n");}}
else{$params{"cutoff"}=1;}
# Check if reserved filenames already exist
if($params{"strand"} eq "0" && -e $params{"in_file"}."_ALL.WIG.gz"){die("Reserved filename already exists ".$params{"in_file"}."_FWD.WIG.gz ...\n");}
if($params{"strand"} eq "1" && -e $params{"in_file"}."_REV.WIG.gz"){die("Reserved filename already exists ".$params{"in_file"}."_FWD.WIG.gz ...\n");}
if($params{"strand"} eq "1" && -e $params{"in_file"}."_FWD.WIG.gz"){die("Reserved filename already exists ".$params{"in_file"}."_FWD.WIG.gz ...\n");}

#======================
# SET MORE VARS
#======================
# Get cwd
my $curr_dir = getcwd;
# Set array with initial chromosome names/numbers according to the uniq headers in the input file
my $line_with_all_chr = `cut -f1 $params{"in_file"} | sort | uniq`;
my @list_of_chr = split(/\n/,$line_with_all_chr);
my %clean_these_up; # Store tmp files to delete later on

#======================
# RUN
#======================
# Run separation on human chr array
&check_input($params{"in_file"});
my $list_of_chr = &separate_per_chr( $params{"in_file"}, \@list_of_chr ); 	#Files are separated on chromosome,new reference list of created files is returned
if($params{"strand"}==1){ 							# If splitting is desired, separate file by strand
	$list_of_chr = &separate_files_by_strand(\@$list_of_chr); 		# Files are split again and new ref to a file list is returned
}
my $list_of_wiggles = &create_wiggles_for_files(\@$list_of_chr); 		# Wiggles are created for submitted files and a new reference to a list of wiggle files returned
my $list_concatted_wiggles = &concat_wiggles(\@$list_of_wiggles); 		# Concatenate wiggles
&compress_files(\@$list_concatted_wiggles); 					# Wiggle files submitted to compress sub for gzip
&cleanup();									# Removal af all files stored in cleanup hash


#======================
# SUBS
#======================
# Check input on IGF format
sub check_input{
#======================
	my $file = shift(@_);
	my $check_this_line = `head -1 $file`; 									# Head the file 
	my @frags = split(/\t/, $check_this_line); 								# Cut it up
	my $read_len = length($frags[4]);
	unless($frags[4] =~ m/[ACGTN]{$read_len}/i){die("Read doesn't look like sequence...\n");}		# Check if the read consists only of allowed characters ACTGN case insensitive for the whole length
	unless($frags[1] !~ m/\D+/){die("START Position not numeric...\n");} 					# Check for digits
	unless($frags[2] !~ m/\D+/){die("END Position not numeric...\n");} 					# Check for digits
	unless($frags[3] =~ m/^[+-]{1}$/){die("Expecting + or - for strand...\n");}				# Check for plus or minus as strand
	unless(length($frags[4]) == length($frags[5])){die("Qual and read are not of equal length...\n");} 	# Check if length is equal for the read and the qual string
}

#======================
# Grep and split file in separate chr-files
# Requires an IGF file and a list of chromosomes to split and sort on
# Returns reference to file list of created files
sub separate_per_chr{
#======================
	my $big_file = shift(@_);
	my $chr_list_ref = shift(@_);
	my @files_for_wiggle;	
	foreach my $chr(@$chr_list_ref){
		print "Grep&Sort for ".$chr." ...\n";
		system(`grep -P "^$chr\t" $big_file | sort -n -k2 >tmp_sorted.${chr}`) || die("Grep&Sort failed\n");
		$clean_these_up{"tmp_sorted.${chr}"}++; # Save prelim files in cleanup hash for later removal
		push(@files_for_wiggle, "tmp_sorted.${chr}");
	}
	return \@files_for_wiggle; # New file_list gets returned
}

#======================
# Separate the tmp files by strand
# Required a list of files to split. Will grep + and - and creaate new files 
# Returns filename array
sub separate_files_by_strand{
#======================
	my $chr_list_ref = shift(@_);
	my @files_for_wiggle;
	foreach my $chr(@$chr_list_ref){
		print "SplitStrands for ".$chr." ...\n";

		system(`grep -P "\t\\+\t" ${chr} >tmp_FWD_splitted_${chr}`) || die("Split + failed\n");
		$clean_these_up{"tmp_FWD_splitted_${chr}"}++; # Save prelim files in cleanup hash for later removal
		push(@files_for_wiggle, "tmp_FWD_splitted_${chr}");
		
		system(`grep -P "\t\-\t" ${chr} >tmp_REV_splitted_${chr}`) || die("Split - failed\n");
		$clean_these_up{"tmp_REV_splitted_${chr}"}++; # Save prelim files in cleanup hash for later removal
		push(@files_for_wiggle, "tmp_REV_splitted_${chr}");
	}
	return \@files_for_wiggle; # New file_list gets returned
}

#======================
# Create wiggle files out of the list of sorted datafiles supplied to this sub
# Requires a cutoff and a file list wich point to igf like files split by chromsome and sorted on start position 
# Possibly splitted by strand
# Returns file list of wiggle files created
# Works with a window approach that will print the window when it reaches a certain size
sub create_wiggles_for_files{
#======================
	my $file_list = shift(@_);
	my @wiggle_files;
	my $cutoff_depth = $params{"cutoff"}; 		# Set cutoff to print depths >= this
	foreach my $file_name(@$file_list){
		my %window; 				# Init the window
		my $window_start;
		my $window_end=0;
		open (IN, "$file_name") || die("Could not open infile $file_name ...\n");
		
		# tmp_sorted.chrY  || tmp_REV_splitted_tmp_sorted.chrY
		# Get chr number
		my $curr_chr = $file_name;
		$curr_chr =~ s/^tmp.+\.//; # Cleanup eveything between tmp and the .
		
		# Pretty filename cleanup some
		my $other_fileinfo = $file_name;
		$other_fileinfo =~ s/^tmp//;
		$other_fileinfo =~ s/_sorted.+//gi;
		$other_fileinfo =~ s/_REV_splitted_tmp/_REV/gi;
		$other_fileinfo =~ s/_FWD_splitted_tmp/_FWD/gi;
		
		my $output_filename .= $curr_chr.$other_fileinfo.".WIG";
		
		open (OUT, ">$output_filename") || die("Could not open infile $file_name ...\n");
		push(@wiggle_files,$output_filename);
		# Wiggle header, more customization can be editted in here
		#print OUT "track type=wiggle_0 name=\"".$curr_chr."\" description=\"Coverage for ".$output_filename."\" visibility=full autoScale=off viewLimits=0.0:25.0 color=200,0,50"."\n";
		print OUT "track type=wiggle_0 name=\"".$output_filename."\" description=\"Coverage for ".$output_filename."\" visibility=full autoScale=on color=200,0,50"."\n";
		print OUT "variableStep chrom=chr$curr_chr\n";
		print "COVERTING $file_name TO WIGGLE $output_filename ...\n";
		while(<IN>){
			chop $_;
			my @frags = split(/\t/, $_);
			# If the read starts before the end of the current 'contig' add up depth
			my $read_len = length($frags[4]);
			if($read_len >999){die("Readlength to long!");} # Adjust minimal left over window size
			if($frags[1] <= $window_end && $window_end - $window_start < 1000000){	# If read starts before current windowend and sizelimit not reached
				for(my $i=$frags[1];$i<=($frags[1]+$read_len-1);$i++){
					$window{ $i } += 1; 					# Add 1 to the depth on this position
					if($window_end < $frags[1]+$read_len-1){
						$window_end = $frags[1]+$read_len-1		# Set window end if this is greater than before we went in for this read
					}
				}
				if ($window_end - $window_start >= 1000000){ 			# Print out and cleanup window if it gets to big
					my $cleanup_counter = 0;
					# === Print and delete previous window
					foreach my $pos( sort {$a<=>$b} keys %window){
						if($window{ $pos } >= $cutoff_depth){
							print OUT $pos."\t".$window{ $pos }."\n";
						}
						delete $window{ $pos };
						$cleanup_counter++;
						$window_start++;
						if($cleanup_counter == 999000){ 		# Stop printing window when 1000 positions are left (if ever the readlentgh>1000 adjust this)
							last;
						}
					}
					# ===
				}
			}
			# Else print and delete the contig so far and init a new one
			else{
				# === Print and delete previous window
				foreach my $pos( sort {$a<=>$b} keys %window){
					if($window{ $pos } >= $cutoff_depth){ 
						print OUT $pos."\t".$window{ $pos }."\n";
					}
					delete $window{ $pos };
				}
				# Init next window from the new startsize of the read not belonging to just printed contig
				$window_start = $frags[1];
				$window_end = $frags[1] + $read_len-1;
				for(my $i=$frags[1];$i<($frags[1]+$read_len);$i++){
					$window{ $i } = 1; # Add 1 to the depth on this position
				}
				# ===	
			}
		}
		# === Print out anything that is still in the window after the last read is parsed
		foreach my $pos( sort {$a<=>$b} keys %window){
			if($window{ $pos } >= $cutoff_depth){
				print OUT $pos."\t".$window{ $pos }."\n";
			}
			delete $window{ $pos };
		}
		# ===
		close(IN);
		close(OUT);
	}
	return \@wiggle_files; #New file_list gets returned
}

#======================
# Concat the indy chromosome wiggle files into one big genome wiggle
# A FWD+REV or ALL file is created for the data and a header is included. Then all data from previous wiggles are grepped and added to this file
# Files are checked for existence before running to avoid contamination of existing files
sub concat_wiggles{
#======================
	my $file_list = shift(@_);
	my @created_files;
	my $fwd_created=0; # Switches used for checking reserved filename when no files are parsed yet.
	my $rev_created=0;
	my $all_created=0;
	my $file_prefix=$params{"in_file"}."_";
	foreach my $file_to_concat(@$file_list){
		print "Concatenating file $file_to_concat ...\n";
		if($file_to_concat =~ /FWD/gi){
			if(!-e $file_prefix."FWD.WIG.gz" && $fwd_created==0){ 			# IF the reserved filename exists AND we haven't parsed anything we go on...
				open(FWD, ">".${file_prefix}."FWD.WIG") || die("Can't open file ${file_prefix}FWD.WIG");
				print FWD "track type=wiggle_0 name=\"${file_prefix}FWD\" description=\"Coverage for ${file_prefix}FWD\" visibility=full autoScale=on color=200,0,50"."\n"; # Create header
				close(FWD);
				$fwd_created=1;							# ... and things to okay.
				push(@created_files,${file_prefix}."FWD.WIG");
			}		
			else{ 									# ELSE we die...
				if($fwd_created==0){die("FWD.WIG.gz already found ...\n");}
			}
			system(`grep -v 'track' $file_to_concat >>${file_prefix}FWD.WIG`); 	# Dump everything in the file exept header line
			$clean_these_up{"$file_to_concat"}++;
		}		
		elsif($file_to_concat =~ /REV/gi){
			if(!-e $file_prefix."REV.WIG.gz" && $rev_created==0){
				open(REV, ">".${file_prefix}."REV.WIG") || die("Can't open file ${file_prefix}REV.WIG");
				print REV "track type=wiggle_0 name=\"${file_prefix}REV\" description=\"Coverage for ${file_prefix}REV\" visibility=full autoScale=on color=200,0,50"."\n";
				close(REV);
				$rev_created=1;
				push(@created_files,${file_prefix}."REV.WIG");
			}
			else{
				if($rev_created==0){die("REV.WIG.gz already found ...\n");}
			}			
			system(`grep -v 'track' $file_to_concat >>${file_prefix}REV.WIG`);
			$clean_these_up{"$file_to_concat"}++;
		}
		else{
			if(!-e $file_prefix."ALL.WIG.gz" && $all_created==0){
				open(ALL, ">".${file_prefix}."ALL.WIG") || die("Can't open file ${file_prefix}ALL.WIG");
				print ALL "track type=wiggle_0 name=\"${file_prefix}ALL\" description=\"Coverage for ${file_prefix}ALL\" visibility=full autoScale=on color=200,0,50"."\n";
				close(ALL);
				$all_created=1;
				push(@created_files,${file_prefix}."ALL.WIG");
			}
			else{
				if($all_created==0){die("ALL.WIG.gz already found ...\n");}
			}			
			system(`grep -v 'track' $file_to_concat >>${file_prefix}ALL.WIG`);
			$clean_these_up{"$file_to_concat"}++;
		}		
	}
	close(FWD);
	close(REV);
	close(ALL);
	return \@created_files;
}

#======================
# Compress the files supplied to this sub
sub compress_files{
#======================
	my $file_list = shift(@_);
	foreach my $file_to_compress(@$file_list){
		print "COMPRESSING $file_to_compress.gz ...\n";
		system(`gzip -f $file_to_compress`) || die("Compression failed ...\n");
		#$clean_these_up{"$file_to_compress"}++; #Not needed, files will be gzipped.
	}
}


#======================
# Cleanup all files stored in cleanup hash
sub cleanup{
#======================
	foreach my $rm_file(keys %clean_these_up){
		system(`rm $rm_file`)|| die("Cleanup failed for $rm_file ...\n");;
	}
}


